"""
"""
name = "cgt"
from sage.all_cmdline import gap
gigs = 30
old_command = gap._Expect__command
s = old_command.index('-o ')
e = old_command.index(' -', s)
gap._Expect__command = gap._Expect__command.replace(gap._Expect__command[s:e], f'-o {gigs}G')
from .examples import *
from .enums import *
from .position_paradigm import PositionParadigmFramework
from .models import Model


"""
"""

class Genome:
    """Optional wrapper for genomes, so that they can be printed/operated on in a more pythonic way."""
    def __init__(self, framework, genome):
        self.framework = framework
        self.generating_object = genome
        try:
            self.canonical_instance = framework.canonical_instance(framework(genome))
        except:
            raise TypeError("Unable to create genome object with given data")


"""
"""

from enum import Enum, auto

class MODEL(Enum):
    one_region_inversions  = "inversions of a single region"
    two_region_inversions  = "inversions of two adjacent regions"
    all_inversions		   = "all inversions"
    one_region_swaps       = "adjacent transpositions"
    two_region_adjacent_transpositions = "transpositions of two adjacent regions"

class TYPE(Enum):
    reg_to_signed_pos = auto()
    pos_to_signed_reg = auto()
    signed_reg_to_pos = auto()
    signed_pos_to_reg = auto()

class FORMAT(Enum):
    formal_sum     = auto()
    equiv_classes  = auto()
    dictionary     = auto()
    only_reps      = auto()

class SET(Enum):
    signed_cycles 	= auto()
    unsigned_cycles = auto()
    one_row			= auto()
    wreath			= auto()
    wreath_s2		= auto()
    
class SYMMETRY(Enum):
    circular = dihedral = D_n              = auto()
    linear = S_2 = C_2 = flip = reflection = auto()

class CLASSES(Enum):
    conjugacy_classes = auto()
    cosets            = auto()
    double_cosets     = auto()

class DISPLAY(Enum):
    arrows  = auto()
    one_row = auto()
    cycles  = auto()

class ALGEBRA(Enum):
    group        = auto()
    genome       = auto()
    genome_class = auto()


"""
"""

from copy import deepcopy
from sage.all_cmdline import SymmetricGroup, floor
import warnings
from . import structures
from .enums import *

def css(cuts_set):
    string = ''.join(sorted([str(cut[0]) for cut in cuts_set]))
    if len(string) != len(cuts_set): print("something went very wrong here...")
    return string

def cuts(framework, sigma):
    sigma = deepcopy(framework.one_row(sigma))
    n = framework.n
    pmN = structures.set_plus_minus(n)
    G = SymmetricGroup(pmN)
    c_string = f'({",".join(str(i) for i in list(range(1, n+1)))})({",".join(str(i) for i in list(range(-n, 0)))})'
    c = G(c_string)
    sigma = list(sigma)
    return set([ 
        tuple(([i+1,c(i+1)])) for i in range(len(sigma)) 
        if (sigma[c(i + 1)-1] != c(sigma[i]))
    ])

def __all_canonical_inversions(framework, num_regions=None):
    if not framework.oriented:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    if framework.symmetry not in {SYMMETRY.circular, SYMMETRY.linear}:
        raise NotImplementedError(f"not implemented for framework with {str(framework.symmetry)} symmetry")
    n = framework.n
    G = framework.genome_group()
    if num_regions == 1:
        return {G(f'({i},-{i})') for i in range(1, n+1)} if framework.oriented else {}
    elif num_regions == 2:
        if framework.oriented:
            perms = {G(f'({i},{-1*(i+1)})({-1*i},{i+1})') for i in range(1,n)}
            if framework.symmetry is not SYMMETRY.linear:
                perms.union({G(f'({n},-1)(-{n},1)')})
        else:
            perms = {G(f'({i},{i+1})') for i in range(1,n)} 
            if framework.symmetry is not SYMMETRY.linear:
                perms.union({G(f'(1,{n})')})
        return perms
    elif num_regions == None: # Return all inversions up to length floor(n/2)
        up_to_length = floor(framework.n/2)
        if framework.oriented: 
            perms = set()
            for permutation in G:
                cycle_type = list(permutation.cycle_type())
                if set(cycle_type)=={1,2} and list(cycle_type).count(2)<=up_to_length and len(cuts(framework, permutation))==2:
                    perms.add(permutation)
            if framework.symmetry is SYMMETRY.linear:
                perms = {perm for perm in perms if not ('1' in str(perm) and str(n) in str(perm))}
            return perms
        else:
            raise NotImplementedError(f"not yet implemented for {framework}")
    else:
        raise NotImplementedError(f"inversions of length {num_regions} not yet implemented")
    
def __two_region_adjacent_transposition_reps(framework):
    if not framework.oriented or framework.symmetry != SYMMETRY.circular:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    G = framework.genome_group()
    return { G('(-2,-1)(1,2)'), G('(-2,1,2,-1)'), G('(-2,-1,2,1)'), G('(-2,2)(-1,1)') }

def double_coset(framework, perm):
    Z = framework.symmetry_group()
    return { d1 * perm * d2 for d1 in Z for d2 in Z }

def conjugacy_class(framework, perm):
    Z = framework.symmetry_group()
    return { d.inverse() * perm * d for d in Z }

def single_coset(framework, perm):
    Z = framework.symmetry_group()
    return { perm * d for d in Z }

def __representatives(framework, set_of_permutations, classes=CLASSES.double_cosets):
    if not framework.oriented:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    Z = framework.symmetry_group()
    perms = {x for x in set_of_permutations}
    cosets = []
    while len(perms):
        perm = perms.pop()
        if classes is CLASSES.double_cosets:
            coset = double_coset(framework, perm)
        elif classes is CLASSES.conjugacy_classes:
            coset = conjugacy_class(framework, perm)
        elif classes is CLASSES.cosets:
            coset = single_coset(framework, perm)
        for element in coset:
            perms.discard(element)
        cosets.append(coset)
    reps = set()
    for coset in cosets:
        reps.add(sorted(list(coset), key=lambda x: (sum(set(x.cycle_type())), list(x.cycle_type()).count(2), str(x)) )[0])
    return reps

def all_inversions_representatives(framework, num_regions=None):
    return __representatives(framework, __all_canonical_inversions(framework, num_regions=num_regions), classes=CLASSES.double_cosets)

def all_adjacent_transpositions_representatives(framework, num_regions=None):
    if num_regions == 2:
        return __two_region_adjacent_transposition_reps(framework)
    else:
        raise NotImplementedError(f"model not yet implemented")



"""
position_paradigm is the most important module in cgt. PositionParadigmFramework keeps track of important objects, can
convert group elements and genomes between different forms, and instances can be passed to functions from other modules so
that they can acces information without it needing to be recreated.
"""

from sage.all_cmdline import QQ, FormalSum, SymmetricGroup, SymmetricGroupRepresentations, gap, UniversalCyclotomicField, Permutations, matrix, QQbar
from sage.combinat.colored_permutations import SignedPermutations
from .enums import *
import numpy as np
import warnings
from copy import deepcopy
from .structures import HyperoctahedralGroup
from scipy.sparse import dok_matrix as dok
from random import choice
from functools import cache

class PositionParadigmFramework:
    """Everything you need for working with genomes under the position paradigm"""

    def __init__(self, num_regions, oriented=True, symmetry=None, genome_type=TYPE.reg_to_signed_pos):
        """Instantiate a system of position paradigm genomes with given set of symmetries and number of regions."""
        if genome_type != TYPE.reg_to_signed_pos:
            raise NotImplementedError("Genome type not yet supported. For now use TYPE.reg_to_signed_pos")
        self.n = num_regions
        self.oriented = oriented
        self.symmetry = symmetry
        
    def __eq__(self, other): # String representation is unique and repr calls str
        return self.__repr__() == other.__repr__()

    def __str__(self):
        string = f"Framework for {str(self.symmetry).replace('SYMMETRY.', '')} genomes"
        string += f" with {self.n}{' oriented' if self.oriented else ''} regions"
        return string

    def __repr__(self):
        return self.__str__() # String representation is unique

    def __call__(self, x):
        """Return an object as a group algebra element if possible"""
        return self.genome_algebra()(self.cycles(x))

    def __hash__(self):
        return self.__str__().__hash__() # String rep is unique, so just hash that!

    def cycles(self, element):
        try: # See if it's already a group element
            return self.genome_group()(element)
        except ValueError:
            pass # Not able to be directly converted, try something else!
        if type(element) is np.ndarray: # If it's a matrix, let's convert it!
            return self._permutation_group_element_from_matrix(element)
        else: # Not a matrix, assume it's a list
            try:
                elt = list(element)
            except TypeError:
                raise ValueError("Invalid data for constructing a permutation")
        if (not self.oriented) or len(elt) == 2*self.n:
            return self.genome_group()(elt)
        elif len(elt) == self.n:
            other_half = list(-int(elt[self.n - i]) for i in range(1, self.n+1))
            return self.genome_group()(elt + other_half)
        else:
            raise ValueError("Invalid data for constructing a permutation")

    def one_row(self, element, as_list=False):
        """Return a given genome instance in one-row notation.

        Args:
            element: the genome instance to represent in one-row notation. Multiple formats accepted.
            as_list: if true, return as a list of images, otherwise return a sage permutation.

        Examples:
            sage: import cgt
            sage: ppf = cgt.PositionParadigmFramework(3)
            sage: ppf.cycles(ppf.one_row('(1,-2)(-1,2)')) == ppf.cycles('(1,-2)(-1,2)')
            True
        """
        elt = self.cycles(element)
        row = list(elt.dict().values())[0:self.n]
        if not as_list:
            if self.oriented:
                row = SignedPermutations(self.n)(row)
            else:
                row = Permutations(self.n)(row)
        return row

    @cache
    def genome_group(self):
        """Return the permutation group containing genome instances."""
        if self.oriented:
            return HyperoctahedralGroup(self.n)
        else:
            return SymmetricGroup(self.n)

    def canonical_instance(self, instance):
        """Return the 'canonical' instance of the genome represented by the permutation if there is one.

        Examples:
            sage: import cgt; ppf = cgt.PositionParadigmFramework(3, symmetry=cgt.SYMMETRY.circular)
            sage: ppf.canonical_instance('(1,-2)(-1,2)')
            [1, 2, -3]
        """
        instance = deepcopy(self.one_row(self.cycles(instance)))
        if self.symmetry in {SYMMETRY.circular, SYMMETRY.linear}:
            f = self.one_row(self.standard_reflection())
            if list(instance)[0] < 0:
                instance = instance*f
            if self.symmetry is SYMMETRY.circular:
                r = self.one_row(self.standard_rotation())
                instance = instance*(r**((self.n-list(instance)[0]+1)%self.n))
            return instance
        else:
            raise NotImplementedError("No canonical form exists in the current framework")

    def random_instance(self, genome=None):
        """Return a random permutation corresponding to a given genome, or a random genome if none is given."""
        if genome:
            return choice(list(genome))[0]
        else:
            return self.genome_group().random_element()

    def random_genome(self, format=FORMAT.formal_sum):
        """Return a random genome"""
        return self.genome(self.random_instance(), format=format)

    @cache
    def symmetry_group(self):
        """Return the symmetry group of the genomes."""
        if self.symmetry == SYMMETRY.circular:
            gens = [self.standard_reflection(), self.standard_rotation()]
        elif self.symmetry == SYMMETRY.linear:
            gens = [self.standard_reflection()]
        else:
            gens = [self.genome_group().one()]
        return self.genome_group().subgroup(gens)

    @cache
    def group_algebra(self):
        """Return the group alegbra, where the group is the group containing genome instances."""
        return self.genome_group().algebra(QQ)

    def symmetry_element(self):
        """Return the symmetry element: a convex sum of elements from symmetry_group()."""
        try:
            return self.z
        except AttributeError:
            self.z = 1/self.symmetry_group().order() * sum(self.group_algebra()(d) for d in self.symmetry_group())
        return self.z

    def num_genomes(self):
        """Return the number of distinct genomes up to symmetries."""
        return self.genome_group().order()/self.symmetry_group().order()

    def _sort_key(self, one_row_perm):
        return str(one_row_perm).replace('-', 'Z')

    def _sort_key_cycles(self, cycles_perm):
        return str(self.one_row(cycles_perm)).replace('-', 'Z')

    def _genome_coset(self, instance):
        coset = set(instance*d for d in self.symmetry_group())
        return sorted([self.one_row(g) for g in coset], key=self._sort_key)

    def _double_coset(self, instance):
        coset = set(d2 * instance * d1 for d1 in self.symmetry_group() for d2 in self.symmetry_group())
        return sorted(coset, key=self._sort_key_cycles)

    def _conjugacy_class(self, instance):
        conj_class = set(d.inverse() * instance * d for d in self.symmetry_group())
        return sorted([self.one_row(g) for g in conj_class], key=self._sort_key)

    def genome_equivalence_classes(self, combine_inverse_classes=False, sort_classes=True):
        """Return double cosets of instances---genomes in the same class will have the same likelihood functions"""
        instances = set(self.genome_group())
        classes = {}
        while len(instances) > 0:
            instance = instances.pop()
            if combine_inverse_classes:
                dcoset = sorted(set(self._double_coset(instance)+self._double_coset(instance.inverse())), key=self._sort_key_cycles)
            else:
                dcoset = self._double_coset(instance)
            instances -= set(dcoset)
            classes[dcoset[0]] = dcoset
        if sort_classes:
            classes = dict(sorted(classes.items(), key=lambda x: self._sort_key_cycles(x[0])))
        return classes
    
    def genome_conjugacy_classes(self, sort_classes=True):
        """Return conjugacy classes of instances"""
        instances = set(self.genome_group())
        classes = {}
        while len(instances) > 0:
            instance = instances.pop()
            conj_class = self._conjugacy_class(instance)
            instances -= set(conj_class)
            classes[conj_class[0]] = conj_class
        if sort_classes:
            classes = dict(sorted(classes.items(), key=lambda x: self._sort_key_cycles(x[0])))
        return classes

    def genome(self, instance, format=None):
        coset = self._genome_coset(instance)
        if format == FORMAT.formal_sum:
            Z = self.symmetry_group()
            A = self.group_algebra()
            return sum(1/Z.order()*A(self.cycles(dx)) for dx in coset)
        else:
            return coset

    def genomes(self, format=FORMAT.dictionary, sort_genomes=True):
        if format not in {FORMAT.dictionary, FORMAT.formal_sum}:
            raise NotImplementedError("Not yet implemented! Convert manually to other formats from FORMAT.dictionary")
        instances = set(self.genome_group())
        genomes = {}
        while len(instances) > 0:
            instance = instances.pop()
            coset = self._genome_coset(instance)
            instances -= set(coset)
            genomes[coset[0]] = coset
        if format == FORMAT.formal_sum:
            Z = self.symmetry_group()
            A = self.group_algebra()
            genomes = { rep : sum(1/Z.order()*A(self.cycles(dx)) for dx in coset) for rep, coset in genomes.items() }
        if sort_genomes:
            genomes = dict(sorted(genomes.items(), key=lambda x: self._sort_key(x[0])))
        return genomes

    def standard_reflection(self):
        """Return a permutation which reflects a genome instance about the center region (n odd), or the center two regions (n even)."""
        if self.oriented:
            string = f'({1},-{self.n})({self.n},-{1})'
            for i in range(1, int(self.n/2)):
                string += f'({i+1}, -{self.n-(i-1)-1})({self.n-(i-1)-1}, -{i+1})'
            if self.n % 2 != 0:
                mid = int(self.n/2) + 1
                string += f'({mid},-{mid})'
        else:
            string = f'(1,{self.n})'
            for i in range(1, int(self.n/2)):
                string += f'({i+1},{self.n-(i-1)-1})'
        return self.genome_group()(string)

    def standard_rotation(self):
        """Return a permutation which rotates positions clockwise by one position."""
        positive = tuple(i for i in range(1,self.n+1))
        string = str(positive)
        if self.oriented:
            negative = tuple(-i for i in positive)
            string += str(negative)
        string = string.replace(',)', ')')
        return self.genome_group()(string)

    def draw_instance(self, instance, shortened=False):
        """Return a one-line drawing of a genome using arrows indicating orientation and elipsis indicating rotational symmetry"""
        permutation = instance.inverse()
        if shortened:
            left_tail, right_head = '\u2758', '\u276D'
            left_head, right_tail = '\u276C', '\u2758'
        else:
            left_tail, right_head = '|-', '->'
            left_head, right_tail = '<-', '-|'
        if self.symmetry in {SYMMETRY.circular, SYMMETRY.linear}:
            string = "..." if self.symmetry is SYMMETRY.circular else ""
            for position in range(1, self.n + 1):
                signed_region = permutation(position)
                if signed_region < 0:
                    region = f"{left_head}{abs(signed_region)}{right_tail}"
                else:
                    region = f"{left_tail}{abs(signed_region)}{right_head}"
                if position == 1:
                    first_region = region
                string += region
            string += f"{first_region}..." if self.symmetry is SYMMETRY.circular else ""
            string = string if self.symmetry is SYMMETRY.circular else string[1:]
            return string if self.oriented else string.replace(right_head, right_tail).replace(left_head, left_tail)
        else:
            raise NotImplementedError(f"Can't draw genome instance with symmetry group {str(self.symmetry)}")

    def _permutation_group_element_from_matrix(self, matrix):
        sigma = []
        for c, col in enumerate(matrix.T):
            image = np.nonzero(col)[0][0]
            sigma.append(matrix[image,c]*(image+1))
        return self.cycles(sigma)

    def matrix(self, element):
        """Return the defining representation matrix. Note that matrix(x)matrix(y) = matrix(yx)"""
        with warnings.catch_warnings(): # Sage uses deprecated objects in to_matrix for coloured permutations. We would like to not be reminded of this.
            warnings.simplefilter("ignore", category=PendingDeprecationWarning)
            return np.array(self.one_row(self.cycles(element)).to_matrix())

    def reg_rep_of_zs(self, model, to_adjacency_matrix=False, sparse=True):
        warnings.warn("Untested! Use at your own risk!")
        # TODO: #14 re-write to directly use model element from the genome algebra
        if self is not model.framework:
            if self != model.framework:
                raise ValueError(f"Current framework and model framework are not the same!")
            else:
                warnings.warn("Current framework and model framework reference different objects! This might cause problems.")
        Z = self.symmetry_group()
        genomes = self.genomes()
        genome_list = list(genomes.values())
        reps = list(genomes.keys()) # Thousands of these
        # Don't need the below line any more
        #model_classes = list({frozenset({ d.inverse() * a * d for d in Z }) for a in model.generating_dictionary.keys()})
        model_generators_cycles = list(model.generating_dictionary.keys()) #[sorted(list(model_class))[0] for model_class in model_classes]
        model_generators = [
            self.one_row(elt) for elt in model_generators_cycles
        ]
        #model_classes = { rep : list(model_classes[r]) for r, rep in enumerate(model_generators) }
        genome_lookup = { rep : r for r, rep in enumerate(reps) }
        num_genomes = len(genomes.keys())
        matrix = dok((num_genomes, num_genomes), dtype=np.float32)
        for g, genome in enumerate(genome_list): # For each genome
            print(f'\rComputing class {g+1} with rep {genome[0]}', end="")
            for permutation in genome:
                for rearrangement in model_generators:
                    coeff = 1
                    if not to_adjacency_matrix:
                        coeff = 1/Z.order() * model.generating_dictionary[model_generators_cycles[model_generators.index(rearrangement)]] # * len(model_classes[rearrangement])
                    matrix[g, genome_lookup[self.canonical_instance(permutation * rearrangement)]] += coeff
        print('...done!')
        return matrix if sparse else matrix.toarray()

    def irreps(self, element=None):
        """
        Return a complete list of pairwise irreducible representations of the genome group.

        Args:
            element: return the image of element for each irreducible representation.

        Returns:
            a complete list of pairwise irreducible representations of the genome group, or a list of matrices if element is not None. 
            Output is cached so that irreps are computed only once per session.
            Matrix entries sit inside the UniversalCyclotomicField and are exact expressions.
        """
        representations = self._cached_irreps()
        if element is not None:
            return [irrep(element) for irrep in representations]
        else:
            return representations

    @cache
    def _cached_irreps(self):
        representations = []
        def irrep_function_factory(irrep, signed):
            def representation(sigma, _irrep=irrep, _signed=signed):
                    result = 0
                    if sigma in self.group_algebra(): # sigma is an algebra element
                        for term in sigma:
                            perm, coeff = term
                            result += coeff * (gap.Image(_irrep, perm) if _signed else _irrep(perm))
                    else: # sigma is a group element
                        result = (gap.Image(_irrep, sigma) if _signed else _irrep(sigma))
                    mat = matrix(UniversalCyclotomicField(), result)
                    return mat.transpose() if _signed else mat
            return representation
        if not self.oriented:
            irreps = SymmetricGroupRepresentations(self.n)
        else:
            irreps = (gap.IrreducibleAffordingRepresentation(character) for character in gap.Irr(self.genome_group()))
        for irrep in irreps:
            representations.append(irrep_function_factory(irrep, self.oriented))
        return representations

    def regular_representation(self, g):
        """Return the regular representation of a single element"""
        warnings.warn("this function is untested! Use at your own risk.", DeprecationWarning)
        return matrix(QQbar, [(self.group_algebra()(g)*self.genome_group()(h)).to_vector(QQbar) for h in self.genome_group()]).transpose()

    def coefficient_in(self, more_terms, fewer_terms):
        """For example, the coefficient of (a/2 + b/2) in x=(a/3 + b/3 + c/3) is 2/3, since x=2/3*(a/2 + b/2) + c/3"""
        coefficients_in_larger_sum = {}
        for term in fewer_terms.terms():
            perm  = self.genome_group()(list(term)[0][0]) # a
            coeff = list(term)[0][1] # 1/2
            coefficients_in_larger_sum[perm] = more_terms.coefficient(perm) * (1/coeff) # 1/3 * 2 = 2/3
        terms = set(coefficients_in_larger_sum.values())
        if len(terms) != 1:
            warnings.warn("Note that extra terms from the smaller sum remain in the larger sum!")
        return min(terms)

    def collect_genome_terms(self, formal_sum, display=DISPLAY.one_row):
        """Return a weighted sum of genomes, represented as [canonical representation]z"""
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            formal_sum = deepcopy(formal_sum)
        terms = formal_sum.terms()
        new_sum = [] # [(coeff, rep), ...]
        while len(terms) > 0:
            term = list(terms[0])[0][0]
            canonical_instance = self.canonical_instance(term)
            genome = self.genome(self.cycles(canonical_instance), format=FORMAT.formal_sum)
            coeff = self.coefficient_in(formal_sum, genome)
            formal_sum -= coeff * genome
            terms = formal_sum.terms()
            if display is DISPLAY.one_row:
                genome_string = str(canonical_instance).replace(' ', '') + 'z'
            elif display is DISPLAY.cycles:
                genome_string = str(self.cycles(canonical_instance)) + 'z'
            elif display is DISPLAY.arrows:
                genome_string = '(' + self.draw_instance(self.cycles(canonical_instance), shortened=True) + ')'
            new_sum.append((coeff, genome_string))
        return FormalSum(new_sum, parent=QQ)



"""
"""

import os
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from .position_paradigm import *
from .enums import *

def draw_genome_instance(framework, instance, show=False):
    """Produce a drawing of a circular genome with n regions"""
    if not framework.oriented:
        raise NotImplementedError("not implemented for linear genomes.")
    try:
        instance = framework(instance)
    except:
        TypeError(f"{instance} does not belong in {str(framework)}")

    matplotlib.rcParams.update({
        "text.usetex": True,
        "font.family": "serif",
        "pgf.texsystem": "pdflatex",
        'pgf.rcfonts': False,
        "font.size": 18,
    })
    plt.style.use('default')

    n = framework.n
    region_labels     = [None for _ in range(1, n+1)]
    orientation_list  = [False for _ in range(1,n+1)]

    for r, p in list(enumerate(framework.one_row(instance, as_list=True))):
        region_labels[abs(p)-1] = r+1
        orientation_list[abs(p)-1] = p<0
    
    region_labels = list(reversed(region_labels))
    orientation_list = list(reversed(orientation_list))

    segment_sizes = [100/n for _ in range(n)]
    segment_colors = ['white' for _ in range(n)]
    
    fig1, ax1 = plt.subplots(figsize=(2,2))
    wedges, texts = ax1.pie(segment_sizes, labeldistance=1.2, colors=segment_colors, 
                            labels=region_labels, radius=1, startangle=90, 
                            wedgeprops={"edgecolor":"0", 'linewidth': 1, 'linestyle': 'solid', 'antialiased': True})
                
    if framework.oriented and orientation_list:
        for i, p in enumerate(wedges):
            ang = (p.theta2 - p.theta1)/2. + p.theta1
            y = np.sin(np.deg2rad(ang))
            x = np.cos(np.deg2rad(ang))
            arrow_length = 0.00001 # We only want the arrow head
            angle = np.deg2rad(ang) + (np.pi/2 if orientation_list[i] else 3*np.pi/2)
            if orientation_list:
                ax1.arrow(x, y, arrow_length*np.cos(angle), arrow_length*np.sin(angle), 
                            overhang = 0.3, head_width=0.12, head_length=0.12, head_starts_at_zero=True, 
                            length_includes_head=True, linewidth=1, facecolor='black', fill=True)

    # Draw inner circle to hide most of the segment edges
    centre_circle = plt.Circle((0,0),0.85,fc='white')
    fig = plt.gcf()
    fig.gca().add_artist(centre_circle)
    ax1.axis('equal') # makes sure it's a circle
    plt.tight_layout()

    if not os.path.exists('_output'):
        os.makedirs('_output')
    plt.savefig(f'_output/{str(instance)}.png', transparent=True)
    plt.savefig(f'_output/{str(instance)}.pdf', transparent=True)
    if show: plt.show()


"""
A place for functions which return important but general algebraic structures requiring only a few parameters
"""

from sage.all_cmdline import SymmetricGroup

def set_plus_minus(n):
    return tuple(range(1,n+1)) + tuple(range(-n, 0))

def HyperoctahedralGroup(n):
    """Return the hyperoctahedral group of order (2^n)n! as a subgroup of SymmetricGroup(2n) with two generators."""
    s_pmn = SymmetricGroup(set_plus_minus(n))
    if n==1: return s_pmn
    h_gens = [
        s_pmn([(1,2),(-1,-2)]),
        s_pmn(tuple(range(1,n+1)) + tuple(range(-1, -(n+1), -1)))
    ]
    return s_pmn.subgroup(h_gens)


"""
"""
from .enums import MODEL, SYMMETRY
from .models import Model
from .position_paradigm import PositionParadigmFramework

def example(n):
    framework = PositionParadigmFramework(n, symmetry=SYMMETRY.circular)
    model = Model.named_model_with_relative_probs(framework, {
        MODEL.one_region_inversions: 2/3, 
        MODEL.two_region_inversions: 1/3
    })
    return framework, model


"""
Implements a number of distance measures for genomes under the position paradigm.
"""

from cgt.enums import ALGEBRA
import numpy as np
import networkx as nx
from sage.all import ComplexDoubleField, UniversalCyclotomicField, matrix, Matrix, real, exp, round, CC
from scipy.optimize import minimize_scalar
from functools import cache

def mles(framework, model, genome_instances):
    """Return maximum likelihood estimates for a set of genome instances under the given model and framework"""
    mles = {}
    for instance in genome_instances:
        mles[instance] = mle(framework, model, instance)
    return mles

def mle(framework, model, genome_instance):
    """Return maximum likelihood estimates for a genome instance under the given model and framework"""
    return maximise(framework, likelihood_function(framework, model, genome_instance, attempt_exact=False))

def maximise(framework, L, max_time=100):
    """Return the time that maximises likelihood function L, using additional information from the framework"""
    limit = 1/framework.num_genomes()
    t_max = minimize_scalar(lambda t: -1*L(t), method='bounded', bounds=(0, max_time))['x']
    mle = t_max if L(t_max)>limit else np.nan
    return mle

def _projection_operators(mat, eigs):
    """Return projection operators for given matrix and its eigenvalues"""
    dim = mat.nrows()
    projections = [matrix.identity(dim) for _ in eigs]
    for e1, eig1 in enumerate(eigs):
        for eig2 in eigs:
            if eig1 != eig2:
                projections[e1] *= (mat-(eig2*matrix.identity(dim)))*(1/(eig1-eig2))
    return projections

def _irreps_of_zs(framework, model, attempt_exact=False):
    """Return a set of matrices---images of zs under each irrep"""
    CDF, UCF = ComplexDoubleField(), UniversalCyclotomicField()
    z = framework.symmetry_element()
    s = model.s_element(in_algebra=ALGEBRA.genome)
    irreps_of_z, irreps_of_s = framework.irreps(z), framework.irreps(s)
    irreps_of_zs = (matrix(UCF, irrep_z*irrep_s) for irrep_z, irrep_s in zip(irreps_of_z, irreps_of_s))
    irreps_of_zs = [Matrix(UCF if attempt_exact else CDF, irrep_zs) for irrep_zs in irreps_of_zs]
    for irrep in irreps_of_zs:
	    irrep.set_immutable()
    return irreps_of_zs

def _eigenvalues(mat, round_to=7, make_real=True, inc_repeated=False, attempt_exact=False, use_numpy=True, bin_eigs=False, tol=10^(-8)):
    """Return all the eigenvalues for a given matrix mat"""
    col = list if inc_repeated else set
    if use_numpy:
        new_mat = np.array(matrix(CC, mat))
        all_eigs = np.linalg.eigvals(new_mat)
    else:
        all_eigs = (eig for eig in mat.eigenvalues())
    if attempt_exact: 
        return sorted(col(all_eigs))
    elif bin_eigs:
        return _bin(sorted(all_eigs), return_bin_size=False, tol=tol)
    else:
        return sorted(col(round(real(eig) if make_real else eig, round_to) for eig in all_eigs))

def _bin(eigenvalues, tol=10^(-8), return_bin_size=False):
    binned_eigenvals = []
    num_eigs = len(eigenvalues) # ??
    e = 0
    eigs = [real(eig) for eig in eigenvalues]
    while e < num_eigs:
        bin=[eigs[e]] # this is the eigenvalue
        while len(bin) < num_eigs-e and abs(eigs[e]-eigs[e+len(bin)]) < tol:
            bin.append(eigs[e+len(bin)])
        average = real(sum(bin))/len(bin)
        binned_eigenvals.append((average, len(bin)) if return_bin_size else average)
        e=e+len(bin)
    return binned_eigenvals

def _eigenvectors(mat, tol=10^(-8)):
    eigen_tuples = sorted(mat.eigenvectors_right())
    binned_eigenvals = _bin([et[0] for et in eigen_tuples], tol=tol, return_bin_size=True)
    # Orthogonalise the eigenvectors                      
    q=0
    eigenvectors = []
    for v in range(len(binned_eigenvals)):
        c=binned_eigenvals[v][1] 
        vec, _ = matrix([eigen_tuples[q+l][1][0].list() for l in range(c)]).gram_schmidt(orthonormal=True)
        eigenvectors.append(vec)
        q=q+c
    return binned_eigenvals, eigenvectors

def _partial_traces_for_genome(framework, instance, irreps, irreps_of_zs, projections, eig_lists, irreps_of_z=None):
    """Return dictionary of partial traces, indexed first by irrep index and then by eigenvalaue"""
    if irreps_of_z is None:
        irreps_of_z = [irrep(framework.symmetry_element()) for irrep in irreps]
    CDF, UCF = ComplexDoubleField(), UniversalCyclotomicField()
    traces = {
        r: {
            eigenvalue: {} for eigenvalue in eig_lists[r]
        } for r in range(len(irreps_of_zs))
    }
    for r, irrep in enumerate(irreps): # Iterate over irreducible representations
        sigd = irreps_of_z[r] * Matrix(CDF, matrix(UCF, irrep(framework.cycles(instance.inverse()))))
        for e, eigenvalue in enumerate(eig_lists[r]):
            traces[r][eigenvalue] = round(real((sigd*projections[r][e]).trace()), 6)
    return traces

def _partial_traces_for_genome_using_eigenvectors(framework, instance, irreps, irreps_of_zs):
    """Return dictionary of partial traces, indexed first by irrep index and then by eigenvalaue"""
    irreps_of_z = [irrep(framework.symmetry_element()) for irrep in irreps]
    eigenvectors_list = [_eigenvectors(irrep_zs) for irrep_zs in irreps_of_zs]
    CDF, UCF = ComplexDoubleField(), UniversalCyclotomicField()
    traces = {
        r: {
            eigenvalue[0]: {} for eigenvalue in eigenvectors_list[r][0]
        } for r in range(len(irreps_of_zs))
    }
    for r, irrep in enumerate(irreps): # Iterate over irreducible representations
        sigd = irreps_of_z[r] * Matrix(CDF, matrix(UCF, irrep(framework.cycles(instance.inverse()))))
        eigenvalues = eigenvectors_list[r][0]
        eigenvector_list = eigenvectors_list[r][1]
        for e, eigenvalue in enumerate(eigenvalues):
            eigenvectors = eigenvector_list[e]
            traces[r][eigenvalue[0]]=round(real((sum([(eigenvectors[m,:].H)*eigenvectors[m,:] for m in range(eigenvalue[1])])*sigd).trace()),6)
    return traces


def likelihood_function(framework, model, genome, attempt_exact=False, use_projections=True):
    """Return the likelihood function for a given genome"""
    instance = genome
    CDF, UCF = ComplexDoubleField(), UniversalCyclotomicField()
    G, Z = framework.genome_group(), framework.symmetry_group()
    irreps = framework.irreps()
    irreps_of_zs = _irreps_of_zs(framework, model, attempt_exact=attempt_exact)
    if use_projections:
        eig_lists = [_eigenvalues(irrep_zs, round_to=7, make_real=True, inc_repeated=False, attempt_exact=attempt_exact) for irrep_zs in irreps_of_zs]
        projections = [_projection_operators(*vals) for vals in zip(irreps_of_zs, eig_lists)]
        traces = _partial_traces_for_genome(framework, instance, irreps, irreps_of_zs, projections, eig_lists)
    else:
        eig_lists = [[x[0] for x in _eigenvectors(irrep_zs)[0]] for irrep_zs in irreps_of_zs]
        traces = _partial_traces_for_genome_using_eigenvectors(framework, instance, irreps, irreps_of_zs)
    dims = [irrep_of_zs.nrows() for irrep_of_zs in irreps_of_zs]
    def likelihood(t):
        ans = 0
        for r, dim in enumerate(dims):
            ans += Z.order()*(exp(-t)/G.order())*dim*sum(exp(eigenvalue*CDF(t)) * traces[r][eigenvalue] for eigenvalue in eig_lists[r])
        return real(ans)
    return likelihood

def min_distance(framework, model, weighted=False):
    """Return dictionary of minimum distances ref->genome, using inverse of model probabilities as weights (or not)"""
    matrix = model.reg_rep_of_zs().toarray()
    if weighted:
        matrix = np.reciprocal(matrix, where=matrix!=0)
    graph = nx.Graph(matrix)
    genome_reps = framework.genomes().keys()
    return { 
        rep : nx.shortest_path_length(graph, source=0, target=r, weight='weighted' if weighted else None) 
        for r, rep in enumerate(genome_reps) 
    }

def MFPT(framework, model, scale_by=1):
    """Return the mean time elapsed rearranging ref->genome where the target is an absorbing state"""
    genomes = framework.genomes()
    P = model.reg_rep_of_zs().toarray()
    P_0 = P.copy()
    for i in range(len(P)):
        P_0[0,i] = 0 
    Z = np.linalg.inv(np.eye(len(P)) - P_0)
    def MFTP_dists():
        return (Z@P_0@Z)[:,0]
    reps = list(genomes.keys())
    MFTP_distances = list(MFTP_dists())
    MFTP_distances = { rep : scale_by * MFTP_distances[r] for r, rep in enumerate(reps) }
    return MFTP_distances


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

def readfile(filename):
    with open(filename) as f:
        return f.read()

setuptools.setup(
    name="cgt",
    version= readfile("VERSION").strip(),
    author="Joshua Stevenson",
    author_email="joshua.stevenson@utas.edu.au",
    description="Tools for representing genomes in sage",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/js51/Circular-genome-tools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
)



from .. import example
import pytest

@pytest.fixture
def example_framework():
    yield example()

def test_one_row_and_cycles_random(example_framework):
    fw, model = example_framework
    instance = fw.random_instance()
    assert instance == fw.cycles(fw.one_row(instance))

def test_matrix_and_cycles_random(example_framework):
    fw, model = example_framework
    instance = fw.random_instance()
    assert instance == fw.cycles(fw.matrix(instance))


"""
"""

from sage.all_cmdline import QQ
from .enums import *
from . import rearrangements
from functools import cache

class Model:
    """Defines a model. A model consists of some collection of permutations and a map from these permutations to probabilities [0,1]"""
    def __init__(self, framework, generating_dictionary):
        """Define a model from a dictionary of single permutations, with their probabilities as the values."""
        self.framework = framework
        self.generating_dictionary = generating_dictionary
        self.names = []
        # TODO: Implement checks for certain model properties, for example time reversibility, symmetry and missing rearrangements

    def __repr__(self):
        return f"Model({str(self.framework)}, {str(self.generating_dictionary)})"

    def __str__(self):
        """Return a descriptive string for the model"""
        # TODO: #21 Make the model able to describe which permutations it allows after it is created
        string = f"Model under the {str(self.framework).lower()}"
        string += f" containing {' and '.join(str(name.value) for name in self.names) if self.names else 'generating perms: ' + str(self.generating_dictionary)}."
        return string

    @classmethod
    def named_model_with_relative_probs(cls, framework, named_model_dictionary):
        if not framework.oriented:
            raise NotImplementedError(f"not yet implemented for {str(framework)}")
        # look through all specified model names and handle them appropriately
        model = {}
        if abs(sum(named_model_dictionary.values()) - 1) > 0.00001: 
            raise ValueError("supplied probabilities do not sum to 1")
        for model_name, relative_prob in named_model_dictionary.items():
            if model_name is MODEL.one_region_inversions:
                gens = rearrangements.all_inversions_representatives(framework, num_regions=1)
                for generator in gens:
                    model[generator] = relative_prob/len(gens)
            if model_name is MODEL.two_region_inversions:
                gens = rearrangements.all_inversions_representatives(framework, num_regions=2)
                for generator in gens:
                    model[generator] = relative_prob/len(gens)
            if model_name is MODEL.all_inversions:
                gens = rearrangements.all_inversions_representatives(framework)
                for generator in gens:
                    model[generator] = relative_prob/len(gens)
            if model_name is MODEL.two_region_adjacent_transpositions:
                gens = rearrangements.all_adjacent_transpositions_representatives(framework, num_regions=2)
                for generator in gens:
                    model[generator] = relative_prob/len(gens)
        model = cls(framework, model)
        model.names += list(named_model_dictionary.keys())
        return model

    @cache
    def reg_rep_of_zs(self):
        """Return the regular representation of zs as comptued by PositionParadigmFramework.reg_rep_zs, but store the sparse result"""
        return self.framework.reg_rep_of_zs(self, sparse=True)

    @cache
    def s_element(self, in_algebra=ALGEBRA.genome):
        if in_algebra not in {ALGEBRA.group, ALGEBRA.genome}:
            raise NotImplementedError(f"Model element for {str(in_algebra)} algebra not yet implemented")
        A = self.framework.group_algebra()
        s = A(0) # Zero element in algebra
        gens_dict = self.generating_dictionary
        gens = {x for x in self.generating_dictionary.keys()}
        while len(gens) > 0:
            gen = gens.pop()
            prob = gens_dict[gen]
            if in_algebra is ALGEBRA.group:
                conj_class = rearrangements.conjugacy_class(self.framework, gen)
            elif in_algebra is ALGEBRA.genome:
                conj_class = {gen}
            for perm in conj_class:
                s += QQ(prob/len(conj_class)) * A(perm)
            gens -= conj_class
        return s